/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl

import scala.annotation.migration
import ObjectParsers._
import ScenarParsers._
import ConversionConfigParsers._

/**
 * This contains the grammar of an internal DSL meant to simplify our lives 
 * when it comes to configuring unit tests scenarios with a framework such as (ie)
 * Mockito.
 */
trait ScenarParsers extends ObjectParsers with ConversionConfigParsers {
  import ScenarParsers._
  /* parses a sequence of instructions separated by a semicolon */
  def scenario:Parser[Seq[Instruction]] = rep1sep(instruction, ";")<~opt(";")
  
  /* parses an instruction */
  def instruction:Parser[Instruction] = declaration | assignment | conversionconfinstr
  
  /* parses a conversion config instruction */
  def conversionconfinstr:Parser[ConversionConfigInstruction] = conversionconf ^^ ConversionConfigInstruction
  
  /* declares an internal identifier (aka a variable that does not exists outside the scenario) */
  def declaration:Parser[Declaration] = internalIdentifier~"="~svalue ^^ { case ref~"="~value => Declaration(ref.name, value)}
  /* parses an assignment (aka the assignment of a value to a given execution chain) */
  def assignment:Parser[Assignment] = memberRef~"="~svalue ^^ {case ref~"="~value => Assignment(ref, value)} 
 
  /* parses a scenario (sequence of instructions) */
  def parseScenario(script:String):Seq[Instruction] = parseAll(scenario, script) match {
    case Success(scn, _) => scn
    case NoSuccess(err, _) => Seq(SyntaxError(err))
  }
}
/** the companion object */
object ScenarParsers {
  sealed case class Instruction
  sealed case class ConversionConfigInstruction(config:ConversionConfig)                          extends Instruction  
  sealed case class Declaration(name:String, value:SLiteral)                                      extends Instruction
  sealed case class Assignment(destination:MemberRef, value:SLiteral)                             extends Instruction
  
  sealed case class SyntaxError(message:String)                                                   extends Instruction
}
