/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.interpreters

import org.mockito.Matchers._
import org.mockito.Mockito._
import xgi.ut.dsl.ObjectParsers._
import xgi.ut.dsl.utl.ReflectionWrapper._
import xgi.ut.dsl.utl.ReflectionUtil

import scala.collection.mutable.HashMap
import scala.collection.JavaConversions._

/**
 * This trait gives the mixed class/object the ability to let
 * a JsonLiteral enter the SStructured+typed world of Java & Scala
 */
trait ToObject extends ConversionConfigInterpreter {
  this:{val context:Any; val variables:HashMap[String, Any]} =>
  /**
   * Let a Literal enter the SStructured+typed world of Java & Scala.
   * A basic type becomes an instance of this basic type.
   * An SArray becomes a java SArray
   * A complextype is transformed into a map where each of the fields is a key.
   */
  def toObject(json:SLiteral):Any = json match {
    // syntax error
    case SSyntaxError(mesg) => Console.err.println(mesg); mesg
    
    // basic types
    case SBasicType(value, None) => value
    case SBasicType(text,  Some(SDeclaredType(typeName))) => convert(text.toString, Class.forName(typeName))
    
    // SArray
    case SArray(elems) => (elems map toObject) toArray
    
    // Reference
    case x:Reference => dereference(x, context)
    
    // Special objects
    case MockObject(SDeclaredType(typeName)) => mock(Class.forName(typeName))
    case WildCardObject(optionalType) => {
      val declType = optionalType getOrElse(SDeclaredType("java.lang.Object"))
      val matcher = declType.name.replace("java.lang.", "").toLowerCase match {
        case "string" => anyString
        case "long" => anyLong
        case "int" => anyInt
        case "integer" => anyInt
        case "short" => anyShort
        case "byte" => anyByte
        case "char" => anyChar
        case "character" => anyChar
        case "float" => anyFloat
        case "double" => anyDouble
        case _ => any(Class.forName(declType.name))
      }
      matcher
    }
    
    // SStructure types
    case SStructure(props, None) => {
      var theMap = new HashMap[String, Any]()
      props map {case SProperty(k, v) => theMap+= k -> toObject(v)}
      theMap
    }
    case SStructure(props, Some(SDeclaredType(typeName))) => {
      val clazz = Class.forName(typeName)
      val instance = instanciate(clazz)
      props foreach { x => 
        val value = toObject(x.value) match {
          case y:HashMap[String, Any] => {
            val valueInstance = instanciate(getFieldType(instance, x.key))
            fromMap(valueInstance, y)
          }
          case y:String => convert(y, getFieldType(instance, x.key))
          case other => other
        }
        setField(instance, x.key, value)
       }
      instance
    } 
    case other => other
  }
  
  /** dereferences a reference */
  def dereference(ref:Reference, currentContext:Any):Any = ref match {
    case InternalIdentifier(name) => variables(name)
    case MemberRef(source, invocations) => {
      val src = dereference(source, currentContext)
      (src /: invocations)((ctx, elem) => dereference(elem, ctx))
    }
    case FieldRef(name) => {
      var externalValue = getFieldValue(currentContext, name)
      // avoid NPE
      if(externalValue==null){
        externalValue = mock(getFieldType(currentContext, name))
        setField(currentContext, name, externalValue)
      }
      externalValue
    }
    case FnInvocation(fname, params) => {
      val args = params map{toObject}
      ReflectionUtil.invoke(currentContext, fname, args, null) // is that ok ?
    }
  }
  
  /** 
   * this will instanciate an object of the class 'clazz' using its default constructor
   * even if it is private.
   */
  def instanciate[A](clazz:Class[A]):A={
    val constructor = clazz.getDeclaredConstructor()
    constructor.setAccessible(true)
    constructor.newInstance()
  }
  
  /** injects all the properties in the given object */
  private[this] def fromMap(target:Any, values:HashMap[String, Any]):Any = {
    values foreach {y => setField(target, y._1, y._2)}
    target
  }
 
}
